import asyncio
import functools
import os
import time

from copy import copy
from prometheus_client import *

start = time.time()

class Timer():
    def __init__(self, name,
                 _text="{}# {}. Elapsed time: {:0.4f} seconds", _logger=print,
                 _prometheus=None, **labels):
        self.name = name
        self._text = _text
        self._logger = _logger

        self.labels = labels
        self.labels["process"] = os.getpid()
        self.labels["process_group"] = os.getppid()

        if _prometheus:  # clone
            self._count, self._failed, self._latency, self._in_progress = _prometheus
        else:
            self._count = Counter(f'operations_{self.name}_count', f'Total count for timer', self.labels)
            self._failed = Counter(f'operations_{self.name}_failed', f'Total failed for timer', self.labels)
            self._latency = Summary(f'operations_{self.name}_latency', f'Latency for timer', self.labels)
            self._in_progress = Gauge(f'operations_{self.name}_in_progress', f'Operations in progress', self.labels)

    def text(self, _text):
        self._text = _text
        return self

    def logger(self, _logger):
        self._logger = _logger
        return self

    @property
    def start(self):
        return self._start_time

    @property
    def elapsed(self):
        return time.perf_counter() - self._start_time

    # use timer object as context manager ('with' syntax)
    def __enter__(self):
        self._start_time = time.perf_counter()
        self._in_progress.labels(**self.labels).inc()
        return self

    def __exit__(self, type, value, traceback):
        global start
        if self._in_progress:
            self._in_progress.labels(**self.labels).dec()
        if self._failed:
            if isinstance(value, BaseException):
                self._failed.labels(**self.labels).inc()
        if self._count:
            self._count.labels(**self.labels).inc()
        if self._latency:
            self._latency.labels(**self.labels).observe(self.elapsed)
        if self.logger and self.name:
            self.logger(self._text.format(self.name, time.time() - start, self.elapsed))

    # use object as decorator
    def __call__(self, func=None, **kwargs):
        if not func:
            registered = set(self.labels.keys()) - {'process', 'process_group'}
            overridden = dict(self.labels)
            for k, v in kwargs.items():
                if not k in registered:
                    raise Exception(f"invalid labels for timer: {kwargs.keys()} [should be from {registered}")
                overridden[k] = v
            return Timer(self.name, self._text, self._logger,
                         (self._count, self._failed, self._latency, self._in_progress),
                         # shouldn't override prometheus metrics
                         **overridden)
        else:
            # use object as decorator
            if asyncio.iscoroutinefunction(func):
                @functools.wraps(func)
                async def wrapper_timer(*args, **kwargs):
                    with copy(self) as timer:
                        # with self as timer:
                        timer._text += " [{}]".format(args)
                        return await func(*args, **kwargs)

                return wrapper_timer
            else:
                @functools.wraps(func)
                def wrapper_timer(*args, **kwargs):
                    with copy(self) as timer:
                        timer._text += " [{}]".format(args)
                        return func(*args, **kwargs)

                return wrapper_timer

    def __str__(self):
        return self.name


if __name__ == '__main__':
    @Timer("testA", label11="value11", label12="value12")
    def a(text):
        print(f"run a('{text}')")
        time.sleep(1)


    timer = Timer("testB", label21="a", label22="c")


    def b(text):
        with timer(label21="value21"):
            print(f"run b('{text}')")
            time.sleep(1)


    a("start11")
    a("start12")
    b("start21")
    b("start22")
    b("start23")

    # influx = Influx(REGISTRY)
    # print(influx.save("count", "sum"))



import inspect
import functools
import time
from timeit import default_timer

from prometheus_client import Counter, Histogram, Gauge, Summary
from quart import request

from lemmings.utils.timer_prom import QuantileSummary


def calc_value(f):
    if not callable(f):
        return lambda x: f
    if inspect.getfullargspec(f).args:
        return lambda x: f(x)
    else:
        return lambda x: f()


class BasePrometheusMetrics(object):
    def __init__(self, app, registry=None):
        self.app = app
        if registry:
            self.registry = registry
        else:
            from prometheus_client import REGISTRY as DEFAULT_REGISTRY
            self.registry = DEFAULT_REGISTRY

    def summary(self, name, description, labels=None, **kwargs):
        return self._track(
            Summary,
            lambda metric, time: metric.observe(time),
            kwargs, name, description, labels
        )

    def gauge(self, name, description, labels=None, **kwargs):
        return self._track(
            Gauge,
            lambda metric, time: metric.dec(),
            kwargs, name, description, labels,
            before=lambda metric: metric.inc()
        )

    def counter(self, name, description, labels=None, **kwargs):
        return self._track(
            Counter,
            lambda metric, time: metric.inc(),
            kwargs, name, description, labels
        )

    def _track(self, metric_type, metric_call, metric_kwargs, name, description, labels, before=None):
        """
        Internal method decorator logic.

        :param metric_type: the type of the metric from the `prometheus_client` library
        :param metric_call: the invocation to execute as a callable with `(metric, time)`
        :param metric_kwargs: additional keyword arguments for creating the metric
        :param name: the name of the metric
        :param description: the description of the metric
        :param labels: a dictionary of `{labelname: callable_or_value}` for labels
        :param before: an optional callable to invoke before executing the
            request handler method accepting the single `metric` argument
        """

        if labels is not None and not isinstance(labels, dict):
            raise TypeError('labels needs to be a dictionary of {labelname: callable}')
        label_names = labels.keys() if labels else tuple()

        label_generator = tuple((key, calc_value(call)) for key, call in labels.items()) if labels else tuple()

        parent_metric = metric_type(
            name, description, labelnames=label_names, registry=self.registry,
            **metric_kwargs
        )

        def get_metric(result):
            if label_names:
                return parent_metric.labels(**{key: call(result) for key, call in label_generator})
            else:
                return parent_metric

        def decorator(f):
            @functools.wraps(f)
            async def func(*args, **kwargs):
                metric = None
                if before:
                    metric = get_metric(None)
                    before(metric)

                start_time = default_timer()
                try:
                    result = await f(*args, **kwargs)
                    if not metric:
                        metric = get_metric(result)
                    total_time = max(default_timer() - start_time, 0)
                    metric_call(metric, time=total_time)
                    return result
                except Exception as ex:
                    raise ex
            return func
        return decorator


class PrometheusMetrics(BasePrometheusMetrics):
    """
    The default metrics include a Histogram for HTTP request latencies
    and number of HTTP requests plus a Counter for the total number
    of HTTP requests.

        @app.route('/<item_type>')
        @metrics.counter('invocation_by_type', 'Number of invocations by type', labels={'item_type': lambda: request.view_args['type']})
        def by_type(item_type):
            pass  # only the counter is collected, not the default metrics

        @app.route('/long-running')
        @metrics.gauge('in_progress', 'Long running requests in progress')
        def long_running():
            pass

        @app.route('/status/<int:status>')
        @metrics.summary('requests_by_status', 'Request latencies by status', labels={'status': lambda r: r.status_code})
        @metrics.histogram('requests_by_status_and_path', 'Request latencies by status and path',
                           labels={'status': lambda r: r.status_code, 'path': lambda: request.path})
        def echo_status(status):
            return 'Status: %s' % status, status

    Label values can be defined as callables:
        - With a single argument that will be the Response object
        - Without an argument, possibly to use with the `request` object
    """

    def __init__(self, app, registry=None):
        super(PrometheusMetrics, self).__init__(app, registry)

        # latency = Summary('http_request_duration_seconds', 'HTTP request duration in seconds',
        #                   ('method', 'path', 'status'), registry=self.registry)
        latency = QuantileSummary('http_request_duration_seconds', 'HTTP request duration in seconds',
                          ('method', 'path', 'status'), registry=self.registry)

        def before_request():
            request.prom_start_time = time.perf_counter()
        def after_request(response):
            if hasattr(request, 'prom_start_time'):
                request.prom_end_time = time.perf_counter()
                total_time = max(time.perf_counter() - request.prom_start_time, 0)
                latency.labels(request.method, request.path, response.status_code).observe(total_time)
            return response
        def teardown_request(exception=None):
            if hasattr(request, 'prom_start_time'):
                total_time = max(time.perf_counter() - request.prom_start_time, 0)
                if not hasattr(request, 'prom_end_time'):
                    latency.labels(request.method, request.path, 500).observe(total_time)
            return
        app.before_request(before_request)
        app.after_request(after_request)
        app.teardown_request(teardown_request)





import datetime
import re


def parse_range(rng):
    for x in rng.split(','):
        x = x.strip()
        if x.isdigit():
            yield int(x)
        elif '-' in x:
            xr = x.split('-')
            yield from range(int(xr[0].strip()), int(xr[1].strip()) + 1)
        elif '..' in x:
            xr = x.split('..')
            yield from range(int(xr[0].strip()), int(xr[1].strip()) + 1)
        else:
            raise ValueError(f"Unknown range specified: {x}")


def _re(variable, re):
    return f"((?P<{variable}>\\d+?)[ ]*({re}))?[ ,]*"


DURATION = re.compile(_re('hours', 'h|hrs?|hours?') + _re('minutes', 'm|mins?') + _re('seconds', 's|sec?'))


def is_num(x):
    return isinstance(x, (int, float, complex)) and not isinstance(x, bool)

def duration(t):
    if isinstance(t, datetime.timedelta):
        return t
    if is_num(t):
        return datetime.timedelta(seconds=t)
    if isinstance(t, str):
        return parse_duration(t)
    return None

def parse_duration(time_str):
    parsed = DURATION.match(time_str.strip())
    if not parsed:
        return None
    params = {}
    for (name, param) in parsed.groupdict().items():
        if param:
            params[name] = int(param)
    if len(params) == 0:
        return None
    return datetime.timedelta(**params)


if __name__ == '__main__':
    print(parse_duration('1hour'))
    print(parse_duration('2 hours , 3 mins'))
    print(parse_duration('5h 3m'))
    print(parse_duration('1x'))



import gc
import logging
import os
import shutil
from http.server import HTTPServer, BaseHTTPRequestHandler
from multiprocessing import Process

import psutil
from prometheus_client import generate_latest, REGISTRY, ProcessCollector, multiprocess, CollectorRegistry, Counter, Gauge

from lemmings.utils.influx import Influx

web_requests = Counter(f'web_requests', f'Latency for incoming web requests', ["url"])
class TestServer(BaseHTTPRequestHandler):
    def do_GET(self):
        web_requests.labels(url="/prometheus").inc()
        self.send_response(200)
        self.send_header('Content-type', 'text/plain')
        self.end_headers()
        self.wfile.write(generate_latest(REGISTRY))

def run_server(server_class=HTTPServer, handler_class=TestServer):
    # reload(prometheus_client)
    multiprocess.MultiProcessCollector(REGISTRY)

    server_address = ('', 8000)
    httpd = server_class(server_address, handler_class)
    try:
        httpd.serve_forever()
    except KeyboardInterrupt:
        pass
    httpd.server_close()
    logging.info("stop server")


ENV = 'prometheus_multiproc_dir'

def gauge(name, comment, fn):
    m = Gauge(name, comment, ["process", "process_group"])
    m.labels(os.getpid(), os.getppid()).set_function(fn)
    return m

def gauge_gen(name, comment, fn):
    m = Gauge(name, comment, ["generation", "process", "process_group"])
    for l in (0, 1, 2):
        m.labels(l, os.getpid(), os.getppid()).set_function(lambda: fn()[l])
    return m

def filter_stat(key):
    return lambda: [x[key] for x in gc.get_stats()]

class ProcessStat:
    def __init__(self):
        self.gc()
        self.mem()

        cpu = gauge('cpu_usage', 'Usage of the CPU in percent', lambda: psutil.cpu_percent(interval=1))
        disk = Gauge('used_disk_space', 'Used disk space in percent', ["partition", "process", "process_group"])
        for partition in psutil.disk_partitions():
            name = partition.mountpoint.replace('/', '_').replace('.', '_')
            disk.labels(name, os.getpid(), os.getppid()).set_function(lambda: psutil.disk_usage(partition.mountpoint).percent)

    def mem(self):
        def total():
            logging.warning(f"!!! {psutil.virtual_memory()}")
            return psutil.virtual_memory().total
        total = gauge('mem_total', 'Total reserver memory', total)
        available = gauge('mem_available', 'Free memory', lambda: psutil.virtual_memory().available)
        used = gauge('mem_used', 'Usage of memory', lambda: psutil.virtual_memory().used)
        percent = gauge('mem_usage', 'Usage of memory in percent', lambda: psutil.virtual_memory().percent)
        logging.warning(f"!!! INIT {psutil.virtual_memory()}")
        return psutil.virtual_memory().total

    def gc(self):
        enabled = gauge('mem_gc_enabled', 'Whether the garbage collector is enabled.', gc.isenabled)
        debug   = gauge('mem_gc_debug', 'The debug flags currently set on the Python GC.', gc.get_debug)
        count   = gauge_gen('mem_gc_count', 'Count of objects tracked by the Python garbage collector, by generation.', gc.get_count)
        threshold = gauge_gen('mem_gc_threshold', 'GC thresholds by generation', gc.get_threshold)
        total     = gauge_gen('mem_gc_collections_total', 'Number of GC collections that occurred by generation', filter_stat('collections'))
        collected = gauge_gen('mem_gc_collected_total', 'Number of garbage collected objects by generation', filter_stat('collected'))
        uncollect = gauge_gen('mem_gc_uncollectables', 'Number of uncollectable objects by generation', filter_stat('uncollectable'))

class Prometheus:
    def __init__(self, registry=REGISTRY,
                 include_process_info=False,
                 shared_dir='./build/prometheus.tmp'):

        if ENV in os.environ:
            self.path = os.environ.get(ENV)
        else:
            logging.info(f"no '{ENV}' in env, '{shared_dir}' will be used instead. smth can go wrong")
            self.path = os.environ.setdefault(ENV, shared_dir)
        shutil.rmtree(self.path, ignore_errors=True)
        os.mkdir(self.path)
        logging.info(f"Prometheus shared directory: {self.path}")

        if not registry:
            registry = CollectorRegistry()
        self.registry = registry
        multiprocess.MultiProcessCollector(self.registry)
        if include_process_info:
            ProcessCollector(registry=self.registry)
        self.influx = Influx(self.registry)
        self.args = []

    def filter(self, *args):
        self.args = args

    def dump_to_influx(self, all=False):
        args = self.args if not all else []
        return self.influx.save(*args)

    def clean(self):
        try:
            shutil.rmtree(self.path)
            logging.debug("temporary prometheus dir cleared")
            self.prom_server.terminate()
            logging.info("prometheus thread terminated")
            self.prom_server.join(5)
            logging.warning("prometheus thread stopped")

        except BaseException as e:
            logging.exception("problem with temporary prometheus dir: ")

    def start_server_process(self):
        self.prom_server = Process(name="PrometheusHttp", target=run_server, args=())
        self.prom_server.start()




import re
from lemmings.utils.utils import ignore_exception

# by default, ansi colors are disabled
RULE = '\\033\\[\d+m'
PATTERN = re.compile(RULE)

def esc(code):
    return f"\033[{code}m"

class Color:
    black, red, green, yellow, blue, magenta, cyan, white = 0, 1, 2, 3, 4, 5, 6, 7

    e = esc(0)  # End of color

    @classmethod
    def enable(cls, f=True):
        global PATTERN, RULE
        PATTERN = re.compile(RULE) if not f else re.compile(r'BAD_COLOR_RULE')

    @classmethod
    def _(cls, typ, color, bright):
        b = ";1" if bright else ""
        return esc(f"{typ}{color}{b}")

    @classmethod
    def _foreground(cls, color, bright=False):
        return cls._(3, color, bright)

    @classmethod
    def _light(cls, color, bright=False):
        return cls._(9, color, bright)

    @classmethod
    def _background(cls, color, bright=False):
        return cls._(4, color, bright)

    @classmethod
    @ignore_exception
    def execute(cls, cmd, msg):
        return cmd + str(msg) + cls.e

    @classmethod
    def bold(cls, msg):
        return Color.execute(esc(1), msg)

    @classmethod
    def cursive(cls, msg):
        return Color.execute(esc(3), msg)

    @classmethod
    def underline(cls, msg):
        return Color.execute(esc(4), msg)

    @classmethod
    def mark(cls, color, msg):
        return Color.execute(Color._foreground(color), msg)

    @classmethod
    def back(cls, color, msg):
        return Color.execute(Color._background(color), msg)

    @classmethod
    @ignore_exception
    def print(cls, color, msg, comment="", prefix=""):
        global PATTERN
        back, fore = cls.list[color % len(cls.list)]
        s = prefix + back + fore + str(msg) + cls.e + str(comment)
        s = PATTERN.sub("", s)
        print(s)

Color.list = [
    (Color._background(Color.black), Color._light(Color.white)),
]
